# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import datetime
import functools
import json
import multiprocessing
import os
import signal
import warnings
import threading
from pathlib import Path

import git
import pygments.lexers
from tqdm import tqdm
from wcmatch import fnmatch

def safe_signal(signum, handler):
    try:
        signal.signal(signum, handler)
    except ValueError:
        # Cannot set signal handler in non-main thread
        pass

# Some filetypes in Pygments are not necessarily computer code, but configuration/documentation. Let's not include those.
IGNORE_PYGMENTS_FILETYPES = [
    "*.json",
    "*.md",
    "*.ps",
    "*.eps",
    "*.txt",
    "*.xml",
    "*.xsl",
    "*.rss",
    "*.xslt",
    "*.xsd",
    "*.wsdl",
    "*.wsf",
    "*.yaml",
    "*.yml",
]

default_filetypes = set()
for _, _, filetypes, _ in pygments.lexers.get_all_lexers():
    default_filetypes.update(filetypes)
default_filetypes.difference_update(IGNORE_PYGMENTS_FILETYPES)


class MiniEntry:
    def __init__(self, entry):
        self.path = entry.path
        self.binsha = entry.binsha


class MiniCommit:
    def __init__(self, commit):
        self.hexsha = commit.hexsha
        self.committed_date = commit.committed_date


def get_top_dir(path):
    return (
        os.path.dirname(path).split("/")[0] + "/"
    )  # Git/GitPython on Windows also returns paths with '/'s


class BlameProc(multiprocessing.Process):
    def __init__(
        self, repo_dir, q, ret_q, run_flag, blame_kwargs, commit2cohort, use_mailmap
    ):
        super().__init__(daemon=True)
        self.repo: git.Repo = git.Repo(repo_dir)
        self.q: multiprocessing.Queue = q
        self.ret_q: multiprocessing.Queue = ret_q
        self.run_flag: multiprocessing.Event = run_flag
        self.blame_kwargs = dict(blame_kwargs)
        self.commit2cohort = commit2cohort  # On Unix systems if process is started via the `fork` method, could make this a copy-on-write variable to save RAM
        self.use_mailmap = use_mailmap

    # Get Blame data for a `file` at `commit`
    def get_file_histogram(self, path, commit):
        h = {}
        try:
            for old_commit, lines in self.repo.blame(commit, path, **self.blame_kwargs):
                cohort = self.commit2cohort.get(old_commit.binsha, "MISSING")
                _, ext = os.path.splitext(path)
                if self.use_mailmap:
                    author_name, author_email = get_mailmap_author_name_email(
                        self.repo, old_commit.author.name, old_commit.author.email
                    )
                else:
                    author_name, author_email = (
                        old_commit.author.name,
                        old_commit.author.email,
                    )
                keys = [
                    ("cohort", cohort),
                    ("ext", ext),
                    ("author", author_name),
                    ("dir", get_top_dir(path)),
                    ("domain", author_email.split("@")[-1]),
                ]

                if old_commit.binsha in self.commit2cohort:
                    keys.append(("sha", old_commit.hexsha))

                for key in keys:
                    h[key] = h.get(key, 0) + len(lines)
        except:
            pass
        return h

    def run(self):
        safe_signal(signal.SIGINT, signal.SIG_IGN)
        try:
            while self.run_flag.wait():
                entry, commit = self.q.get()
                if not commit:
                    return
                self.ret_q.put((entry, self.get_file_histogram(entry, commit)))
        except:
            raise


class BlameDriver:
    def __init__(
        self,
        repo_dir,
        proc_count,
        last_file_y,
        cur_y,
        blame_kwargs,
        commit2cohort,
        use_mailmap,
        quiet,
    ):
        self.repo_dir = repo_dir
        self.proc_count = proc_count
        self.q = multiprocessing.Queue()
        self.ret_q = multiprocessing.Queue()
        self.run_flag = multiprocessing.Event()
        self.run_flag.set()
        self.last_file_y = last_file_y
        self.cur_y = cur_y
        self.blame_kwargs = blame_kwargs
        self.commit2cohort = commit2cohort
        self.use_mailmap = use_mailmap
        self.quiet = quiet
        self.proc_pool = []
        self.spawn_process(self.proc_count)

    def spawn_process(self, spawn_only=False):
        n = self.proc_count - len(self.proc_pool)
        if n == 0:
            return
        if n < 0:
            return None if spawn_only else self._despawn_process(-n)
        if not self.quiet:
            print("\n\nStarting up processes: ", end="")
        for i in range(n):
            self.proc_pool.append(
                BlameProc(
                    self.repo_dir,
                    self.q,
                    self.ret_q,
                    self.run_flag,
                    self.blame_kwargs,
                    self.commit2cohort,
                    self.use_mailmap,
                )
            )
            self.proc_pool[-1].start()
            if not self.quiet:
                print(
                    ("" if i == 0 else ", ") + self.proc_pool[-1].name,
                    end="\n" if i == n - 1 else "",
                )

    def _despawn_process(self, n):
        for i in range(n):
            self.q.put((None, None))

        print("\n")
        while True:
            print("\rShutting down processes: ", end="")
            killed_processes = 0
            for idx, proc in enumerate(self.proc_pool):
                if proc.is_alive():
                    continue
                else:
                    print(
                        ("" if killed_processes == 0 else ", ") + proc.name,
                        end="\n" if killed_processes == n - 1 else "",
                    )
                    killed_processes += 1
            if killed_processes >= n:
                for proc in self.proc_pool:
                    if not proc.is_alive():
                        proc.join()
                self.proc_pool = [proc for proc in self.proc_pool if proc.is_alive()]
                return

    def fetch(self, commit, check_entries, bar):
        self.spawn_process()
        processed_entries = 0
        total_entries = len(check_entries)

        for entry in check_entries:
            self.q.put((entry.path, commit.hexsha))

        while processed_entries < total_entries:
            path, file_y = self.ret_q.get()

            for key_tuple, file_locs in file_y.items():
                self.cur_y[key_tuple] = self.cur_y.get(key_tuple, 0) + file_locs
            self.last_file_y[path] = file_y

            processed_entries += 1
            self.run_flag.wait()
            bar.update()

        return self.cur_y

    def pause(self):
        self.run_flag.clear()

    def resume(self):
        self.run_flag.set()


def analyze(
    repo_dir,
    cohortfm="%Y",
    interval=7 * 24 * 60 * 60,
    ignore=[],
    only=[],
    outdir=".",
    branch="master",
    all_filetypes=False,
    ignore_whitespace=False,
    procs=2,
    quiet=False,
    opt=False,
    progress_callback=None,
):
    use_mailmap = (Path(repo_dir) / ".mailmap").exists()
    repo = git.Repo(repo_dir)
    blame_kwargs = {}
    if ignore_whitespace:
        blame_kwargs["w"] = True
    master_commits = []  # only stores a subset
    commit2cohort = {}
    curve_key_tuples = set()  # Keys of each curve that will be tracked
    tqdm_args = {
        "smoothing": 0.025,  # Exponential smoothing is still rather jumpy, a tiny number will do
        "disable": quiet,
        "dynamic_ncols": True,
    }

    if outdir and not os.path.exists(outdir):
        os.makedirs(outdir)

    # Check if specified branch exists
    try:
        repo.git.show_ref("refs/heads/{:s}".format(branch), verify=True)
    except git.exc.GitCommandError:
        default_branch = repo.active_branch.name
        warnings.warn(
            "Requested branch: '{:s}' does not exist. Falling back to default branch: '{:s}'".format(
                branch, default_branch
            )
        )

        branch = default_branch

    if not quiet and repo.git.version_info < (2, 31, 0):
        print(
            "Old Git version {:d}.{:d}.{:d} detected. There are optimizations available in version 2.31.0 which speed up performance".format(
                *repo.git.version_info
            )
        )

    if opt:
        if not quiet:
            print(
                "Generating git commit-graph... If you wish, this file is deletable later at .git/objects/info"
            )
        repo.git.execute(
            ["git", "commit-graph", "write", "--changed-paths"]
        )  # repo.git.commit_graph('write --changed-paths') doesn't work for some reason

    desc = "{:<55s}".format("Listing all commits")
    for i, commit in enumerate(tqdm(
        repo.iter_commits(branch), desc=desc, unit=" Commits", **tqdm_args
    )):
        if progress_callback:
            progress_callback(0, f"Listing commits: {i}")
        cohort = datetime.datetime.utcfromtimestamp(commit.committed_date).strftime(
            cohortfm
        )
        commit2cohort[commit.binsha] = cohort
        curve_key_tuples.add(("cohort", cohort))
        if use_mailmap:
            author_name, author_email = get_mailmap_author_name_email(
                repo, commit.author.name, commit.author.email
            )
        else:
            author_name, author_email = commit.author.name, commit.author.email
        curve_key_tuples.add(("author", author_name))
        curve_key_tuples.add(("domain", author_email.split("@")[-1]))

    desc = "{:<55s}".format("Backtracking the master branch")
    with tqdm(desc=desc, unit=" Commits", **tqdm_args) as bar:
        commit = repo.head.commit
        last_date = None
        while True:
            if last_date is None or commit.committed_date < last_date - interval:
                master_commits.append(commit)
                last_date = commit.committed_date
            bar.update()
            if progress_callback:
                progress_callback(0, f"Backtracking branch: {len(master_commits)} commits found")
            if not commit.parents:
                break
            commit = commit.parents[0]
        del commit

    if ignore and not only:
        only = ["**"]  # stupid fix
    def_ft_str = "+({:s})".format("|".join(default_filetypes))
    path_match_str = "{:s}|!+({:s})".format("|".join(only), "|".join(ignore))
    path_match_zero = len(only) == 0 and len(ignore) == 0
    ok_entry_paths = dict()
    all_entries = []

    def entry_path_ok(path):
        # All this matching is slow so let's cache it
        if path not in ok_entry_paths:
            ok_entry_paths[path] = (
                all_filetypes
                or fnmatch.fnmatch(
                    os.path.split(path)[-1], def_ft_str, flags=fnmatch.EXTMATCH
                )
            ) and (
                path_match_zero
                or fnmatch.fnmatch(
                    path,
                    path_match_str,
                    flags=fnmatch.NEGATE | fnmatch.EXTMATCH | fnmatch.SPLIT,
                )
            )
        return ok_entry_paths[path]

    def get_entries(commit):
        tmp = [
            MiniEntry(entry)
            for entry in commit.tree.traverse()
            if entry.type == "blob" and entry_path_ok(entry.path)
        ]
        all_entries.append(tmp)
        return tmp

    master_commits = master_commits[::-1]  # Reverse it so it's chnological ascending
    entries_total = 0
    desc = "{:<55s}".format("Discovering entries & caching filenames")
    with tqdm(
        desc="{:<55s}".format("Entries Discovered"),
        unit=" Entries",
        position=1,
        **tqdm_args,
    ) as bar:
        for i, commit in enumerate(
            tqdm(master_commits, desc=desc, unit=" Commits", position=0, **tqdm_args)
        ):
            for entry in get_entries(commit):
                entries_total += 1
                _, ext = os.path.splitext(entry.path)
                curve_key_tuples.add(("ext", ext))
                curve_key_tuples.add(("dir", get_top_dir(entry.path)))
                bar.update()
                if progress_callback:
                    progress_callback(0, f"Discovering entries: {entries_total}")
            master_commits[i] = MiniCommit(
                commit
            )  # Might have cached the entries, we don't want that

    # We don't need these anymore, let GC Cleanup
    del repo
    del ok_entry_paths
    del commit
    # End GC Cleanup

    curves = {}  # multiple y axis, in the form key_tuple: Array[y-axis points]
    ts = []  # x axis
    last_file_y = (
        {}
    )  # Contributions of each individual file to each individual curve, when the file was last seen
    cur_y = {}  # Sum of all contributions between files towards each individual curve
    blamer = BlameDriver(
        repo_dir,
        procs,
        last_file_y,
        cur_y,
        blame_kwargs,
        commit2cohort,
        use_mailmap,
        quiet,
    )
    commit_history = (
        {}
    )  # How many lines of a commit (by SHA) still exist at a given time
    last_file_hash = {}  # File SHAs when they were last seen

    # Allow script to be paused and process count to change
    def handler(a, b):
        try:
            blamer.pause()
            print("\n\nProcess paused")
            x = int(
                input(
                    "0. Exit\n1. Continue\n2. Modify process count\nSelect an option: "
                )
            )

            if x == 1:
                return blamer.resume()
            elif x == 2:
                x = int(
                    input(
                        "\n\nCurrent Processes: {:d}\nNew Setting: ".format(
                            blamer.proc_count
                        )
                    )
                )
                if x > 0:
                    blamer.proc_count = x
                    blamer.spawn_process(spawn_only=True)
                return blamer.resume()
            os._exit(1)  # sys.exit() does weird things
        except:
            pass
        handler(None, None)

    if not quiet:
        safe_signal(signal.SIGINT, handler)

    desc = "{:<55s}".format(
        "Analyzing commit history with {:d} processes".format(procs)
    )
    with tqdm(
        desc="{:<55s}".format("Entries Processed"),
        total=entries_total,
        unit=" Entries",
        position=1,
        maxinterval=1,
        miniters=100,
        **tqdm_args,
    ) as bar:
        cbar = tqdm(master_commits, desc=desc, unit=" Commits", position=0, **tqdm_args)
        for i, commit in enumerate(cbar):
            if progress_callback:
                progress_callback(i / len(master_commits), f"Analyzing commit {i+1}/{len(master_commits)}")
            t = datetime.datetime.utcfromtimestamp(commit.committed_date)
            ts.append(t)  # x axis

            # START: Fast diff, to reduce no. of files checked via blame.
            # File hashes are checked against previous iteration
            entries = all_entries.pop(
                0
            )  # all_entries grows smaller as curves grows larger

            check_entries = []
            cur_file_hash = {}
            for entry in entries:
                cur_file_hash[entry.path] = entry.binsha
                if entry.path in last_file_hash:
                    if last_file_hash[entry.path] != entry.binsha:  # Modified file
                        for key_tuple, count in last_file_y[entry.path].items():
                            cur_y[key_tuple] -= count
                        check_entries.append(entry)
                    else:  # Identical file
                        bar.update()
                    del last_file_hash[
                        entry.path
                    ]  # Identical/Modified file removed, leaving deleted files behind
                else:  # Newly added file
                    check_entries.append(entry)
            for deleted_path in last_file_hash.keys():  # Deleted files
                for key_tuple, count in last_file_y[deleted_path].items():
                    cur_y[key_tuple] -= count
            last_file_hash = cur_file_hash
            # END: Fast diff

            # Multiprocess blame checker, updates cur_y & last_file_y
            blamer.fetch(commit, check_entries, bar)
            cbar.set_description(
                "{:<55s}".format(
                    "Analyzing commit history with {:d} processes".format(
                        len(blamer.proc_pool)
                    )
                ),
                False,
            )

            for key_tuple, count in cur_y.items():
                key_category, key = key_tuple
                if key_category == "sha":
                    commit_history.setdefault(key, []).append(
                        (commit.committed_date, count)
                    )

            for key_tuple in curve_key_tuples:
                curves.setdefault(key_tuple, []).append(cur_y.get(key_tuple, 0))

    safe_signal(signal.SIGINT, signal.default_int_handler)

    results = {}

    def get_data(key_type, label_fmt=lambda x: x):
        key_items = sorted(k for t, k in curve_key_tuples if t == key_type)
        return {
            "y": [curves[(key_type, key_item)] for key_item in key_items],
            "ts": [t.isoformat() for t in ts],
            "labels": [label_fmt(key_item) for key_item in key_items],
        }

    results["cohorts"] = get_data("cohort", lambda c: "Code added in %s" % c)
    results["exts"] = get_data("ext")
    results["authors"] = get_data("author")
    results["dirs"] = get_data("dir")
    results["domains"] = get_data("domain")
    results["survival"] = commit_history

    if outdir:
        for key in ["cohorts", "exts", "authors", "dirs", "domains"]:
            fn = os.path.join(outdir, f"{key}.json")
            if not quiet:
                print("Writing data to %s" % fn)
            with open(fn, "w") as f:
                json.dump(results[key], f)
        
        # Survival data
        fn = os.path.join(outdir, "survival.json")
        if not quiet:
            print("Writing survival data to %s" % fn)
        with open(fn, "w") as f:
            json.dump(commit_history, f)
    
    return results


@functools.lru_cache(maxsize=None)
def get_mailmap_author_name_email(repo, author_name, author_email):
    pre_mailmap_author_email = f"{author_name} <{author_email}>"
    mail_mapped_author_email: str = repo.git.check_mailmap(pre_mailmap_author_email)
    mailmap_name, mailmap_email = mail_mapped_author_email[:-1].split(" <", maxsplit=1)
    return mailmap_name, mailmap_email


def analyze_cmdline():
    parser = argparse.ArgumentParser(description="Analyze git repo")
    parser.add_argument(
        "--cohortfm",
        default="%Y",
        type=str,
        help='A Python datetime format string such as "%%Y" for creating cohorts (default: %(default)s)',
    )
    parser.add_argument(
        "--interval",
        default=7 * 24 * 60 * 60,
        type=int,
        help="Min difference between commits to analyze (default: %(default)ss)",
    )
    parser.add_argument(
        "--ignore",
        default=[],
        action="append",
        help="File patterns that should be ignored (can provide multiple, will each subtract independently). Uses glob syntax and generally needs to be shell escaped. For instance, to ignore a subdirectory `foo/`, run `git-of-theseus . --ignore 'foo/**'`.",
    )
    parser.add_argument(
        "--only",
        default=[],
        action="append",
        help="File patterns that can match. Multiple can be provided. If at least one is provided, every file has to match at least one. Uses glob syntax and typically has to be shell escaped. In order to analytize a subdirectory `bar/`, run `git-of-theseus . --only 'bar/**'`",
    )
    parser.add_argument(
        "--outdir",
        default=".",
        help="Output directory to store results (default: %(default)s)",
    )
    parser.add_argument(
        "--branch",
        default="master",
        type=str,
        help="Branch to track (default: %(default)s)",
    )
    parser.add_argument(
        "--ignore-whitespace",
        default=[],
        action="store_true",
        help="Ignore whitespace changes when running git blame.",
    )
    parser.add_argument(
        "--all-filetypes",
        action="store_true",
        help="Include all files (if not set then will only analyze %s"
        % ",".join(default_filetypes),
    )
    parser.add_argument(
        "--quiet",
        action="store_true",
        help="Disable all console output (default: %(default)s)",
    )
    parser.add_argument(
        "--procs",
        default=2,
        type=int,
        help="Number of processes to use. There is a point of diminishing returns, and RAM may become an issue on large repos (default: %(default)s)",
    )
    parser.add_argument(
        "--opt",
        action="store_true",
        help="Generates git commit-graph; Improves performance at the cost of some (~80KB/kCommit) disk space (default: %(default)s)",
    )
    parser.add_argument("repo_dir")
    kwargs = vars(parser.parse_args())

    try:
        analyze(**kwargs)
    except KeyboardInterrupt:
        exit(1)
    except:
        raise


if __name__ == "__main__":
    analyze_cmdline()



# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools, numpy


def generate_n_colors(n):
    vs = numpy.linspace(0.4, 0.9, 6)
    colors = [(0.9, 0.4, 0.4)]

    def euclidean(a, b):
        return sum((x - y) ** 2 for x, y in zip(a, b))

    while len(colors) < n:
        new_color = max(
            itertools.product(vs, vs, vs),
            key=lambda a: min(euclidean(a, b) for b in colors),
        )
        colors.append(new_color)
    return colors



# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib

matplotlib.use("Agg")

import argparse
import collections
import json
import math
import os
import sys

import numpy
from matplotlib import pyplot


def survival_plot(
    input_fns, exp_fit=False, display=False, outfile=None, years=5, title=None
):
    all_deltas = []
    YEAR = 365.25 * 24 * 60 * 60
    fig = pyplot.figure(figsize=(13, 8))
    pyplot.style.use("ggplot")

    for fn_or_data in input_fns:
        if isinstance(fn_or_data, str):
            print("reading %s" % fn_or_data)
            commit_history = json.load(open(fn_or_data))
            fn = fn_or_data
        else:
            commit_history = fn_or_data
            fn = "data"

        print("counting %d commits" % len(commit_history))
        deltas = collections.defaultdict(lambda: numpy.zeros(2))
        total_n = 0
        for commit, history in commit_history.items():
            t0, orig_count = history[0]
            total_n += orig_count
            last_count = orig_count
            for t, count in history[1:]:
                deltas[t - t0] += (count - last_count, 0)
                last_count = count
            deltas[history[-1][0] - t0] += (-last_count, -orig_count)

        all_deltas.append((total_n, deltas))
        print("adding %d deltas..." % len(deltas))
        total_k = total_n
        P = 1.0
        xs = []
        ys = []
        for t in sorted(deltas.keys()):
            delta_k, delta_n = deltas[t]
            xs.append(t / YEAR)
            ys.append(100.0 * P)
            P *= 1 + delta_k / total_n
            total_k += delta_k
            total_n += delta_n
            if P < 0.05:
                break

        print("plotting...")
        if exp_fit:
            pyplot.plot(xs, ys, color="darkgray")
        else:
            parts = os.path.split(fn)
            pyplot.plot(xs, ys, label=(len(parts) > 1 and parts[-2] or None))

    def fit(k):
        loss = 0.0
        for total_n, deltas in all_deltas:
            total_k = total_n
            P = 1.0
            for t in sorted(deltas.keys()):
                delta_k, delta_n = deltas[t]
                pred = total_n * math.exp(-k * t / YEAR)
                loss += (total_n * P - pred) ** 2
                P *= 1 + delta_k / total_n
                total_k += delta_k
                total_n += delta_n
        print(k, loss)
        return loss

    if exp_fit:
        try:
            import scipy.optimize
        except ImportError:
            sys.exit("Scipy is a required dependency when using the --exp-fit flag")

        print("fitting exponential function")
        k = scipy.optimize.fmin(fit, 0.5, maxiter=50)[0]
        ts = numpy.linspace(0, years, 1000)
        ys = [100.0 * math.exp(-k * t) for t in ts]
        pyplot.plot(
            ts,
            ys,
            color="red",
            label="Exponential fit, half-life = %.2f years" % (math.log(2) / k),
        )

    pyplot.xlabel("Years")
    pyplot.ylabel("%")
    pyplot.xlim([0, years])
    pyplot.ylim([0, 100])
    pyplot.title("% of lines still present in code after n years")
    pyplot.legend()
    pyplot.tight_layout()

    if title:
        pyplot.text(0.5, 0.5, title, transform=pyplot.gca().transAxes, 
                    fontsize=40, color='gray', alpha=0.3,
                    ha='center', va='center', rotation=30)

    if outfile:
        print("Writing output to %s" % outfile)
        pyplot.savefig(outfile)
    
    if display:
        pyplot.show()
    
    return fig


def survival_plot_cmdline():
    parser = argparse.ArgumentParser(description="Plot survival plot")
    parser.add_argument("--exp-fit", action="store_true", help="Plot exponential fit")
    parser.add_argument("--display", action="store_true", help="Display plot")
    parser.add_argument(
        "--outfile",
        default="survival_plot.png",
        type=str,
        help="Output file to store results (default: %(default)s)",
    )
    parser.add_argument(
        "--years",
        type=float,
        default=5,
        help="Number of years on x axis (default: %(default)s)",
    )
    parser.add_argument("input_fns", nargs="*")
    kwargs = vars(parser.parse_args())

    survival_plot(**kwargs)


if __name__ == "__main__":
    survival_plot_cmdline()



import sys
import os
import subprocess

def main():
    # Get the directory of the current file
    cmd_dir = os.path.dirname(os.path.abspath(__file__))
    app_path = os.path.join(cmd_dir, "app.py")
    
    # Always use the current working directory
    repo_path = os.path.abspath(os.getcwd())
    
    # Run streamlit
    # We pass the repo_path as an argument to the streamlit script
    try:
        subprocess.run([
            sys.executable, "-m", "streamlit", "run", 
            app_path, 
            "--browser.gatherUsageStats", "false",
            "--", repo_path
        ])
    except KeyboardInterrupt:
        # Exit gracefully on Ctrl+C
        sys.exit(0)

if __name__ == "__main__":
    main()



import plotly.graph_objects as go
import plotly.express as px
import numpy as np
import dateutil.parser
import collections
import math
import os
from .utils import generate_n_colors

# Harmonious, professional color palette (Modern & Muted)
# Inspired by Tableau 20 and modern UI systems
PREMIUM_PALETTE = [
    "#4E79A7", "#A0CBE8", "#F28E2B", "#FFBE7D", "#59A14F", 
    "#8CD17D", "#B6992D", "#F1CE63", "#499894", "#86BCB6",
    "#E15759", "#FF9D9A", "#79706E", "#BAB0AC", "#D37295", 
    "#FABFD2", "#B07AA1", "#D4A1D2", "#9D7660", "#D7B5A6"
]

def _process_stack_line_data(data, max_n=20, normalize=False):
    if not isinstance(data, dict):
        import json
        data = json.load(open(data))

    y = np.array(data["y"])
    labels = data["labels"]
    ts = [dateutil.parser.parse(t) for t in data["ts"]]

    if y.shape[0] > max_n:
        js = sorted(range(len(labels)), key=lambda j: max(y[j]), reverse=True)
        other_indices = js[max_n:]
        if other_indices:
            other_sum = np.sum([y[j] for j in other_indices], axis=0)
            top_js = sorted(js[:max_n], key=lambda j: labels[j])
            y = np.array([y[j] for j in top_js] + [other_sum])
            labels = [labels[j] for j in top_js] + ["other"]
    
    y_sums = np.sum(y, axis=0)
    y_sums[y_sums == 0] = 1.0

    if normalize:
        y = 100.0 * y / y_sums

    return ts, y, labels

def plotly_stack_plot(data, max_n=20, normalize=False, title=None):
    ts, y, labels = _process_stack_line_data(data, max_n, normalize)
    fig = go.Figure()
    
    for i, label in enumerate(labels):
        color = PREMIUM_PALETTE[i % len(PREMIUM_PALETTE)]
        fig.add_trace(go.Scatter(
            x=ts, 
            y=y[i], 
            mode='lines',
            name=label,
            stackgroup='one', 
            line=dict(width=0.5, color='rgba(255,255,255,0.3)'), 
            fillcolor=color,
            hoverinfo='x+y+name'
        ))

    fig.update_layout(
        title=dict(text=title, x=0.5, font=dict(size=20)) if title else None,
        yaxis=dict(
            title="Share of LoC (%)" if normalize else "Lines of Code",
            range=[0, 100.1] if normalize else None,
            gridcolor='rgba(128,128,128,0.2)'
        ),
        xaxis=dict(title="Date", gridcolor='rgba(128,128,128,0.2)'),
        hovermode="x unified",
        margin=dict(l=20, r=20, t=60, b=120),
        legend=dict(orientation="h", yanchor="top", y=-0.25, xanchor="center", x=0.5),
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)'
    )
    return fig

def plotly_line_plot(data, max_n=20, normalize=False, title=None):
    ts, y, labels = _process_stack_line_data(data, max_n, normalize)
    fig = go.Figure()

    for i, label in enumerate(labels):
         fig.add_trace(go.Scatter(
            x=ts, 
            y=y[i], 
            mode='lines',
            name=label,
            line=dict(width=2.5, color=PREMIUM_PALETTE[i % len(PREMIUM_PALETTE)])
        ))

    fig.update_layout(
        title=dict(text=title, x=0.5, font=dict(size=20)) if title else None,
        yaxis=dict(
            title="Share of LoC (%)" if normalize else "Lines of Code",
            range=[0, 100.1] if normalize else None,
            gridcolor='rgba(128,128,128,0.2)'
        ),
        xaxis=dict(title="Date", gridcolor='rgba(128,128,128,0.2)'),
        hovermode="x unified",
        margin=dict(l=20, r=20, t=60, b=120),
        legend=dict(orientation="h", yanchor="top", y=-0.25, xanchor="center", x=0.5),
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)'
    )
    return fig

def plotly_survival_plot(commit_history, exp_fit=False, years=5, title=None):
    deltas = collections.defaultdict(lambda: np.zeros(2))
    total_n = 0
    YEAR = 365.25 * 24 * 60 * 60
    
    for commit, history in commit_history.items():
        t0, orig_count = history[0]
        total_n += orig_count
        last_count = orig_count
        for t, count in history[1:]:
            deltas[t - t0] += (count - last_count, 0)
            last_count = count
        deltas[history[-1][0] - t0] += (-last_count, -orig_count)

    P = 1.0
    xs, ys = [], []
    sorted_times = sorted(deltas.keys())
    
    for t in sorted_times:
        delta_k, delta_n = deltas[t]
        xs.append(t / YEAR)
        ys.append(100.0 * P)
        if total_n > 0:
             P *= 1 + delta_k / total_n
        total_n += delta_n
        if P < 0.05: break
            
    fig = go.Figure()
    fig.add_trace(go.Scatter(
        x=xs, y=ys,
        mode='lines',
        name='Survival Rate',
        line=dict(color=PREMIUM_PALETTE[0], width=3)
    ))

    if exp_fit:
        try:
            import scipy.optimize
            def fit(k):
                loss, curr_total_n = 0.0, sum(h[0][1] for h in commit_history.values())
                P_fit, curr_total_n_fit = 1.0, curr_total_n
                for t in sorted_times:
                    delta_k, delta_n = deltas[t]
                    pred = curr_total_n_fit * math.exp(-k * t / YEAR)
                    loss += (curr_total_n_fit * P_fit - pred) ** 2
                    if curr_total_n_fit > 0: P_fit *= 1 + delta_k / curr_total_n_fit
                    curr_total_n_fit += delta_n
                return loss
            k_opt = scipy.optimize.fmin(fit, 0.5, maxiter=50, disp=False)[0]
            ts_fit = np.linspace(0, years, 100)
            ys_fit = [100.0 * math.exp(-k_opt * t) for t in ts_fit]
            half_life = math.log(2) / k_opt
            fig.add_trace(go.Scatter(
                x=ts_fit, y=ys_fit,
                mode='lines',
                name=f"Exp. Fit (Half-life: {half_life:.2f} yrs)",
                line=dict(color=PREMIUM_PALETTE[10], dash='dash', width=2)
            ))
        except ImportError: pass

    fig.update_layout(
        title=dict(text=title, x=0.5) if title else None,
        yaxis=dict(title="Lines still present (%)", range=[0, 105], gridcolor='rgba(128,128,128,0.2)'),
        xaxis=dict(title="Years", range=[0, years], gridcolor='rgba(128,128,128,0.2)'),
        hovermode="x unified",
        margin=dict(l=20, r=20, t=50, b=120),
        legend=dict(orientation="h", yanchor="top", y=-0.25, xanchor="center", x=0.5),
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)'
    )
    return fig

def plotly_bar_plot(data, max_n=20, title=None):
    _, y, labels = _process_stack_line_data(data, max_n, normalize=False)
    latest_values = [row[-1] for row in y]
    indices = sorted(range(len(labels)), key=lambda i: latest_values[i], reverse=True)
    sorted_labels = [labels[i] for i in indices]
    sorted_values = [latest_values[i] for i in indices]
    
    fig = go.Figure(go.Bar(
        x=sorted_labels,
        y=sorted_values,
        marker=dict(
            color=sorted_values,
            colorscale=[[i/(len(PREMIUM_PALETTE)-1), c] for i, c in enumerate(PREMIUM_PALETTE)],
            showscale=False
        )
    ))

    fig.update_layout(
        title=dict(text=f"{title} (Latest)" if title else "Latest Distribution", x=0.5),
        yaxis=dict(title="Lines of Code", gridcolor='rgba(128,128,128,0.2)'),
        xaxis=dict(title=""),
        margin=dict(l=20, r=20, t=50, b=100),
        plot_bgcolor='rgba(0,0,0,0)',
        paper_bgcolor='rgba(0,0,0,0)'
    )
    return fig



import streamlit as st
import os
import tempfile
import shutil
try:
    from git_of_theseus.analyze import analyze
    from git_of_theseus.plotly_plots import plotly_stack_plot, plotly_line_plot, plotly_survival_plot, plotly_bar_plot
except ImportError:
    from analyze import analyze
    from plotly_plots import plotly_stack_plot, plotly_line_plot, plotly_survival_plot, plotly_bar_plot

st.set_page_config(page_title="Git of Theseus Dash", layout="wide")

# GitHub Link in Sidebar
st.sidebar.markdown(
    """
    <div style="display: flex; align-items: center; margin-bottom: 20px;">
        <img src="https://github.githubassets.com/images/modules/logos_page/GitHub-Mark.png" width="30" style="margin-right: 10px;">
        <a href="https://github.com/onewesong/better-git-of-theseus" target="_blank" style="text-decoration: none; color: inherit; font-weight: bold;">
            better-git-of-theseus
        </a>
    </div>
    """,
    unsafe_allow_html=True
)

st.title("📊 Git of Theseus - Repository Analysis")

import sys

# Sidebar Configuration
st.sidebar.header("Configuration")

with st.sidebar.expander("📖 How to use", expanded=False):
    st.markdown("""
    **Better Git of Theseus** is a tool to analyze the evolution of Git repositories.
    
    ### Plots Explained:
    - **Stack Plot**: Shows code growth over time, broken down by cohort (when code was added).
    - **Line Plot**: Shows trends across different dimensions (Author, Extension, etc.).
    - **Distribution**: Shows the **current** distribution (Who contributed most, which file types are dominant).
    - **Survival Plot**: Estimates how long a line of code typically lasts before being modified or deleted.
    
    ### Tips:
    - **Cohort Format**: `%Y` (Yearly) and `%Y-%m` (Monthly) are recommended.
    - **Mailmap**: Use a `.mailmap` file in the repo root to resolve duplicate author names.
    """)

default_repo = "."
if len(sys.argv) > 1:
    default_repo = sys.argv[1]

repo_path = default_repo
repo_name = os.path.basename(os.path.abspath(repo_path))
st.sidebar.markdown(f"**Repo:** {repo_name}")

# Fetch branches for the selectbox
try:
    import git
    repo = git.Repo(repo_path)
    # Get local branches
    branches = [h.name for h in repo.heads]
    
    # Try to determine the best default branch (active one, or master/main)
    try:
        current_active = repo.active_branch.name
    except:
        current_active = "master"
        
    if current_active in branches:
        branches.remove(current_active)
    
    options = [current_active] + sorted(branches)
    branch = st.sidebar.selectbox("Branch", options=options)
except Exception as e:
    # Fallback if git repo access fails
    branch = st.sidebar.text_input("Branch", value="master")

with st.sidebar.expander("Analysis Parameters"):
    cohortfm = st.text_input(
        "Cohort Format", 
        value="%Y",
        help="Python strftime format string. Common options:\n\n"
             "- `%Y`: Year (e.g., 2023)\n"
             "- `%Y-%m`: Month (e.g., 2023-01)\n"
             "- `%Y-W%W`: Week (e.g., 2023-W01)\n"
             "- `%Y-%m-%d`: Day"
    )
    interval = st.number_input(
        "Analysis Interval (seconds)", 
        value=7 * 24 * 60 * 60,
        help="The time step between data points. Default is 604800s (7 days). Larger values are faster; smaller values result in smoother curves."
    )
    st.caption(f"Current resolution: {interval / 86400:.1f} days")
    
    procs = st.number_input(
        "Parallel Processes", 
        value=2, 
        min_value=1,
        help="Number of concurrent processes. Increase to speed up analysis on multi-core CPUs, but note it increases RAM usage."
    )
    ignore = st.text_area(
        "Ignore Patterns",
        help="Glob patterns to ignore (comma separated), e.g.: 'tests/**, *.md'"
    ).split(",")
    ignore = [i.strip() for i in ignore if i.strip()]

@st.cache_data(show_spinner=False)
def run_analysis(repo_path, branch, cohortfm, interval, procs, ignore):
    # This is a dummy wrapper for the cached function
    # Because we need the callback to update the UI, we can't easily cache the callback-enabled version
    # However, we can cache the final result.
    return analyze(
        repo_path,
        cohortfm=cohortfm,
        interval=interval,
        ignore=ignore,
        outdir=None,
        branch=branch,
        procs=procs,
        quiet=True
    )

def run_analysis_with_progress(repo_path, branch, cohortfm, interval, procs, ignore):
    progress_bar = st.progress(0)
    status_text = st.empty()
    
    def progress_callback(progress, text):
        progress_bar.progress(progress)
        status_text.text(text)
    
    results = analyze(
        repo_path,
        cohortfm=cohortfm,
        interval=interval,
        ignore=ignore,
        outdir=None,
        branch=branch,
        procs=procs,
        quiet=True,
        progress_callback=progress_callback
    )
    
    progress_bar.empty()
    status_text.empty()
    return results

# State management for analysis results
if 'analysis_results' not in st.session_state:
    st.session_state.analysis_results = None

if st.sidebar.button("🚀 Run Analysis") or (len(sys.argv) > 1 and st.session_state.analysis_results is None):
    try:
        st.session_state.analysis_results = run_analysis_with_progress(
            repo_path, branch, cohortfm, interval, procs, ignore
        )
        st.success("Analysis completed!")
    except Exception as e:
        st.error(f"Analysis failed: {e}")
        st.session_state.analysis_results = None

# Main View
if st.session_state.analysis_results:
    results = st.session_state.analysis_results
    tab1, tab2, tab3, tab4 = st.tabs(["Stack Plot", "Line Plot", "Distribution", "Survival Plot"])

    with tab1:
        st.header("Stack Plot")
        col1, col2 = st.columns([1, 3])
        with col1:
            source_map = {
                "Cohorts": "cohorts",
                "Authors": "authors",
                "Extensions": "exts",
                "Directories": "dirs",
                "Domains": "domains"
            }
            data_source_label = st.selectbox("Data Source", list(source_map.keys()), key="stack_source")
            data_key = source_map[data_source_label]
            normalize = st.checkbox("Normalize to 100%", value=False, key="stack_norm")
            max_n = st.slider("Max Series", 5, 50, 20, key="stack_max_n")
        with col2:
            project_name = os.path.basename(os.path.abspath(repo_path))
            data = results.get(data_key)
            if data:
                fig = plotly_stack_plot(data, normalize=normalize, max_n=max_n, title=project_name)
                st.plotly_chart(fig, width="stretch")
            else:
                st.warning(f"Data for {data_source_label} not found.")

    with tab2:
        st.header("Line Plot")
        col1, col2 = st.columns([1, 3])
        with col1:
            data_source_label_line = st.selectbox("Data Source", list(source_map.keys()), key="line_source")
            data_key_line = source_map[data_source_label_line]
            normalize_line = st.checkbox("Normalize to 100%", value=False, key="line_norm")
            max_n_line = st.slider("Max Series", 5, 50, 20, key="line_max_n")
        with col2:
            project_name = os.path.basename(os.path.abspath(repo_path))
            data_line = results.get(data_key_line)
            if data_line:
                fig = plotly_line_plot(data_line, normalize=normalize_line, max_n=max_n_line, title=project_name)
                st.plotly_chart(fig, width="stretch")
            else:
                st.warning(f"Data for {data_source_label_line} not found.")

    with tab3:
        st.header("Latest Distribution")
        col1, col2 = st.columns([1, 3])
        with col1:
            data_source_label_bar = st.selectbox("Data Source", list(source_map.keys()), key="bar_source")
            data_key_bar = source_map[data_source_label_bar]
            max_n_bar = st.slider("Max Series", 5, 100, 30, key="bar_max_n")
        with col2:
            project_name = os.path.basename(os.path.abspath(repo_path))
            data_bar = results.get(data_key_bar)
            if data_bar:
                fig = plotly_bar_plot(data_bar, max_n=max_n_bar, title=f"{project_name} - {data_source_label_bar}")
                st.plotly_chart(fig, width="stretch")
            else:
                st.warning(f"Data for {data_source_label_bar} not found.")

    with tab4:
        st.header("Survival Plot")
        col1, col2 = st.columns([1, 3])
        with col1:
            exp_fit = st.checkbox("Exponential Fit", value=False)
            years = st.slider("Years", 1, 20, 5)
        with col2:
            project_name = os.path.basename(os.path.abspath(repo_path))
            survival_data = results.get("survival")
            if survival_data:
                fig = plotly_survival_plot(survival_data, exp_fit=exp_fit, years=years, title=project_name)
                st.plotly_chart(fig, width="stretch")
            else:
                st.warning("Survival data not found.")

else:
    st.info("👈 Enter a repository path and click 'Run Analysis' to get started.")



# -*- coding: utf-8 -*-
#
# Copyright 2016 Erik Bernhardsson
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib

matplotlib.use("Agg")

import argparse, dateutil.parser, json, numpy, sys
from matplotlib import pyplot

from .utils import generate_n_colors


def stack_plot(
    input_fn_or_data, display=False, outfile=None, max_n=20, normalize=False, title=None
):
    if isinstance(input_fn_or_data, str):
        data = json.load(open(input_fn_or_data)) 
    else:
        data = input_fn_or_data
    y = numpy.array(data["y"])
    if y.shape[0] > max_n:
        js = sorted(range(len(data["labels"])), key=lambda j: max(y[j]), reverse=True)
        other_sum = numpy.sum(y[j] for j in js[max_n:])
        top_js = sorted(js[:max_n], key=lambda j: data["labels"][j])
        y = numpy.array([y[j] for j in top_js] + [other_sum])
        labels = [data["labels"][j] for j in top_js] + ["other"]
    else:
        labels = data["labels"]
    if normalize:
        y = 100.0 * numpy.array(y) / numpy.sum(y, axis=0)
    fig = pyplot.figure(figsize=(16, 12), dpi=120)
    pyplot.style.use("ggplot")
    ts = [dateutil.parser.parse(t) for t in data["ts"]]
    colors = generate_n_colors(len(labels))
    pyplot.stackplot(ts, numpy.array(y), labels=labels, colors=colors)
    pyplot.legend(loc=2)
    if normalize:
        pyplot.ylabel("Share of lines of code (%)")
        pyplot.ylim([0, 100])
    else:
        pyplot.ylabel("Lines of code")

    if title:
        pyplot.text(0.5, 0.5, title, transform=pyplot.gca().transAxes, 
                    fontsize=40, color='gray', alpha=0.3,
                    ha='center', va='center', rotation=30)
    
    if outfile:
        print("Writing output to %s" % outfile)
        pyplot.savefig(outfile)
    
    pyplot.tight_layout()
    if display:
        pyplot.show()
    
    return fig


def stack_plot_cmdline():
    parser = argparse.ArgumentParser(description="Plot stack plot")
    parser.add_argument("--display", action="store_true", help="Display plot")
    parser.add_argument(
        "--outfile",
        default="stack_plot.png",
        type=str,
        help="Output file to store results (default: %(default)s)",
    )
    parser.add_argument(
        "--max-n",
        default=20,
        type=int,
        help='Max number of dataseries (will roll everything else into "other") (default: %(default)s)',
    )
    parser.add_argument(
        "--normalize", action="store_true", help="Normalize the plot to 100%%"
    )
    parser.add_argument("input_fn")
    kwargs = vars(parser.parse_args())

    stack_plot(**kwargs)


if __name__ == "__main__":
    stack_plot_cmdline()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="better-git-of-theseus",
    version="0.6.0",
    description="Plot stats on Git repositories with interactive Plotly charts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Erik Bernhardsson",
    author_email="mail@erikbern.com",
    url="https://github.com/onewesong/better-git-of-theseus",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "gitpython",
        "numpy",
        "tqdm",
        "wcmatch",
        "pygments",
        "plotly",
        "streamlit",
        "python-dateutil",
        "scipy",
        "matplotlib"
    ],
    entry_points={
        "console_scripts": [
            "better-git-of-theseus=git_of_theseus.cmd:main",
        ]
    },
)



from git_of_theseus.analyze import analyze, analyze_cmdline
from git_of_theseus.survival_plot import survival_plot, survival_plot_cmdline
from git_of_theseus.stack_plot import stack_plot, stack_plot_cmdline
from git_of_theseus.line_plot import line_plot, line_plot_cmdline
